from starlette.authentication import AuthenticationError


class InvalidCredentials(AuthenticationError):
    def __init__(
        self,
        message: str = "Could not authenticate with the provided credentials",
    ):
        super().__init__(message)



import base64
import binascii
import typing

from starlette import authentication as auth
from starlette.requests import HTTPConnection

from ..datatypes import AuthResult
from ..exceptions import InvalidCredentials


class AuthBackend(auth.AuthenticationBackend):
    async def authenticate(self, conn: HTTPConnection) -> AuthResult:
        raise NotImplementedError


class _BaseSchemeAuth(AuthBackend):
    scheme: str

    def get_credentials(self, conn: HTTPConnection) -> typing.Optional[str]:
        if "Authorization" not in conn.headers:
            return None

        authorization = conn.headers.get("Authorization")
        scheme, _, credentials = authorization.partition(" ")
        if scheme.lower() != self.scheme.lower():
            return None

        return credentials

    def parse_credentials(self, credentials: str) -> typing.List[str]:
        return [credentials]

    verify: typing.Callable[
        ..., typing.Awaitable[typing.Optional[auth.BaseUser]]
    ]

    async def authenticate(self, conn: HTTPConnection):
        credentials = self.get_credentials(conn)
        if credentials is None:
            return None

        parts = self.parse_credentials(credentials)
        user = await self.verify(*parts)
        if user is None:
            raise InvalidCredentials

        return auth.AuthCredentials(["authenticated"]), user


class BaseBasicAuth(_BaseSchemeAuth):
    scheme = "Basic"

    def parse_credentials(self, credentials: str) -> typing.List[str]:
        try:
            decoded_credentials = base64.b64decode(credentials).decode("ascii")
        except (ValueError, UnicodeDecodeError, binascii.Error):
            raise InvalidCredentials

        try:
            username, password = decoded_credentials.split(":")
        except ValueError:
            raise InvalidCredentials

        return [username, password]

    async def find_user(self, username: str) -> typing.Optional[auth.BaseUser]:
        raise NotImplementedError

    async def verify_password(self, user: auth.BaseUser, password: str) -> bool:
        raise NotImplementedError

    async def verify(
        self, username: str, password: str
    ) -> typing.Optional[auth.BaseUser]:
        user = await self.find_user(username=username)

        if user is None:
            return None

        valid = await self.verify_password(user, password)
        if not valid:
            return None

        return user


class BaseTokenAuth(_BaseSchemeAuth):
    scheme = "Token"

    def parse_credentials(self, credentials: str) -> typing.List[str]:
        token = credentials
        return [token]

    async def verify(self, token: str) -> typing.Optional[auth.BaseUser]:
        raise NotImplementedError



import typing

from starlette import authentication as auth
from starlette.requests import HTTPConnection

from .base.backends import AuthBackend
from .datatypes import AuthResult


class MultiAuth(AuthBackend):
    def __init__(self, backends: typing.List[AuthBackend]):
        self.backends = backends

    async def authenticate(self, conn: HTTPConnection) -> AuthResult:
        for backend in self.backends:
            try:
                auth_result = await backend.authenticate(conn)
            except auth.AuthenticationError as exc:
                raise exc from None

            if auth_result is None:
                continue

            return auth_result

        return None



from pathlib import Path

from setuptools import find_packages, setup

with open(str(Path(".") / "README.md"), encoding="utf-8") as f:
    README = f.read()

PASSWORDS_REQUIREMENTS = ["passlib>=1.7, <2"]

setup(
    name="starlette-auth-toolkit",
    version="0.5.0",
    author="Florimond Manca",
    author_email="florimond.manca@gmail.com",
    description=(
        "Authentication backends and helpers for "
        "Starlette-based ASGI apps and frameworks"
    ),
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/florimondmanca/starlette-auth-toolkit.git",
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    zip_safe=False,
    install_requires=["starlette >= 0.11"],
    extras_require={
        "passwords": PASSWORDS_REQUIREMENTS,
        "dev": [
            # Tests
            "pytest",
            "pytest-asyncio",
            # Test client
            "requests",
            # Web server
            "uvicorn",
            # Optional password backends
            "argon2-cffi",
            "bcrypt",
            # orm integration
            "orm",
            "databases[sqlite]",
            # Code style
            "black",
            "pylint",
            # Release
            "bumpversion",
        ],
    },
    python_requires=">=3.6",
    # license and classifier list:
    # https://pypi.org/pypi?%3Aaction=list_classifiers
    license="License :: OSI Approved :: MIT License",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ],
)



import secrets
import string
import typing

from starlette.concurrency import run_in_threadpool

try:
    import passlib.hash as _hashers
    from passlib.ifc import PasswordHash
except ImportError:  # pragma: no cover
    _hashers = None  # type: ignore
    PasswordHash = None  # type: ignore


def generate_random_string(size: int = 32) -> str:
    return "".join(
        secrets.choice(generate_random_string.alphabet) for _ in range(size)
    )


generate_random_string.alphabet = string.ascii_letters + string.digits


class BaseHasher:
    async def make(self, secret: str) -> str:
        return await run_in_threadpool(self.make_sync, secret)

    async def verify(self, secret: str, hashed: str) -> bool:
        return await run_in_threadpool(self.verify_sync, secret, hashed)

    def make_sync(self, secret: str) -> str:
        raise NotImplementedError

    def verify_sync(self, secret: str, hashed: str) -> bool:
        raise NotImplementedError

    def needs_update(
        self, hashed: str  # pylint: disable=unused-argument
    ) -> bool:
        return False


class Hasher(BaseHasher):
    def __init__(self, algorithm: str):
        assert (
            _hashers is not None
        ), "'passlib' must be installed to use password hashers"
        self.algorithm = algorithm
        try:
            self._hasher: PasswordHash = getattr(_hashers, algorithm)
        except AttributeError as exc:
            raise ValueError(f"unknown algorithm: {algorithm}") from exc

    def make_sync(self, secret: str) -> str:
        return self._hasher.hash(secret)

    def verify_sync(self, secret: str, hashed: str) -> bool:
        return self._hasher.verify(secret, hashed)

    def needs_update(self, hashed: str) -> bool:
        return self._hasher.needs_update(hashed)

    def identify(self, hashed: str) -> bool:
        return self._hasher.identify(hashed)


class PBKDF2Hasher(Hasher):
    def __init__(self):
        super().__init__("pbkdf2_sha256")


# Requires `bcrypt`
class BCryptHasher(Hasher):
    def __init__(self):
        super().__init__("bcrypt")


# Requires `argon2-cffi`
class Argon2Hasher(Hasher):
    def __init__(self):
        super().__init__("argon2")


class CryptHasher(Hasher):
    def __init__(self):
        super().__init__("sha256_crypt")


class MultiHasher(BaseHasher):
    _dummy_secret = "dummysecret"

    def __init__(self, hashers: typing.List[Hasher]):
        if not hashers:
            raise ValueError("'hashers' should contain at least one hasher")
        self.hashers = hashers
        self._needs_update = None
        self._dummy_hash = self.make_sync(self._dummy_secret)

    @property
    def default_hasher(self) -> BaseHasher:
        return self.hashers[0]

    def make_sync(self, secret: str) -> str:
        return self.default_hasher.make_sync(secret)

    def verify_sync(self, secret: str, hashed: str) -> bool:
        self._needs_update = None

        for hasher in self.hashers:
            if hasher.identify(hashed):
                if self._needs_update is None:
                    self._needs_update = hasher.needs_update(hashed)
                return hasher.verify_sync(secret, hashed)
            self._needs_update = True

        # Verify dummy password to reduce vulnerability to timing attacks.
        self.default_hasher.verify_sync(self._dummy_secret, self._dummy_hash)

        return False

    def needs_update(self, hashed: str) -> bool:
        if self._needs_update is None:
            raise RuntimeError(
                "cannot call '.needs_update()' before calling '.verify()'"
            )
        needs_update, self._needs_update = self._needs_update, None
        return needs_update



import inspect
import typing

import orm

from ..base.backends import BaseBasicAuth
from ..cryptography import BaseHasher

_UserModel = typing.Type[orm.Model]
_User = orm.Model


class ModelBasicAuth(BaseBasicAuth):
    _model: _UserModel

    def __init__(
        self,
        model: typing.Union[_UserModel, typing.Callable[[], _UserModel]],
        *,
        hasher: BaseHasher,
        password_field: str = "password",
    ):
        if inspect.isclass(model) and issubclass(model, orm.Model):
            self._get_model = lambda: model
        else:
            assert inspect.isfunction(model)
            self._get_model = model
        self.hasher = hasher
        self.password_field = password_field

    @property
    def model(self) -> _UserModel:
        try:
            return self._model
        except AttributeError:
            self._model = self._get_model()
            return self._model

    async def find_user(self, username: str) -> typing.Optional[_User]:
        try:
            return await self.model.objects.get(username=username)
        except orm.NoMatch:
            return None

    async def verify_password(self, user: _User, password: str):
        password_hash = getattr(user, self.password_field)
        valid = await self.hasher.verify(password, password_hash)

        if not valid:
            return False

        if self.hasher.needs_update(password_hash):
            new_hash = await self.hasher.make(password)
            await user.update(**{self.password_field: new_hash})

        return True
